from clangast.data_structures.node import BaseNode as Node
from clangast.data_structures.tree import BaseTree as Tree

from clangast.algorithms.parsing import (
    get_translation_unit_index, parse_translation_unit,
    parse_translation_unit_generator
)

from clangast.algorithms.__utility__ import __init_clang__

__default_path_to_library = "/usr/lib/libclang.so"

__init_clang__(__default_path_to_library)

__all__ = [
    "Node", "Tree",
    "get_translation_unit_index", "parse_translation_unit",
    "parse_translation_unit_generator"
]



from functools import partial as __partial__
from clangast.algorithms.__utility__ import (
    __all_as_kinds_generator__, __cursor_kind__,
    __not_none__, __token_kind__
)


def get_available_cursor_kinds():
    return filter(__not_none__,
                  __cursor_kind__.__dict__["_kinds"])


def get_available_token_kinds():
    return filter(__not_none__,
                  __token_kind__.__dict__["_kinds"])


all_function_declarations_generator = __partial__(__all_as_kinds_generator__,
                                                  [__cursor_kind__.FUNCTION_DECL])



from clangast.data_structures.tree import BaseTree as Tree
from clangast.algorithms.utility import (
    all_function_declarations_generator
)
from clangast.algorithms.__parsing__ import (
    __parse_function__, __cursor_index__
)

from typing import List, Any, Generator

TranslationUnit = __cursor_index__.TranslationUnit


def get_translation_unit_index(path_to_file: str, **kwargs) -> TranslationUnit:
    return __cursor_index__.Index.create().parse(path_to_file, **kwargs)


def parse_translation_unit_generator(unit: TranslationUnit) -> Generator[Tree, Any, None]:
    for function_declaration in all_function_declarations_generator(unit.cursor):
        yield __parse_function__(function_declaration)


def parse_translation_unit(unit: TranslationUnit) -> List[Tree]:
    return list(parse_translation_unit_generator(unit))



from clangast.data_structures.node import BaseNode
from typing import List, Union, DefaultDict
from collections import defaultdict
from functools import wraps

import sys


class BaseTree(object):
    reverse_index_type = DefaultDict[BaseNode, List[List[int]]]

    @staticmethod
    def __not_empty_tree__(method):
        @wraps(method)
        def __check__(self, *args, **kwargs):
            if self.empty():
                raise RuntimeError("Empty tree")
            method(self, *args, **kwargs)

        return __check__

    @staticmethod
    def __child_indexing_after__(method):
        @wraps(method)
        def __perform__(self, element, *args, **kwargs):
            result = method(self, element, *args, **kwargs)
            if result == 1:
                self.__child_indexing__(element, self.__last_inserted_path)
            return result

        return __perform__

    def __init__(self, root: BaseNode = None):
        self._root = root
        self._counter = 0
        self._reverse_index: reverse_index_type = defaultdict(list)

        if self._root is not None:
            self._counter += 1

        self.__last_inserted_path: List[int] = []

    def __len__(self) -> int:
        return self._counter

    @__not_empty_tree__.__get__(object)
    def __merger_reverse_index_modified_paths__(self, another_reverse_index: reverse_index_type,
                                                additional_path: List[int]):
        for element, paths in another_reverse_index.items():
            modified_paths = []
            for path in paths:
                modified_paths.append(path + additional_path)
            self.__add_paths__(element, modified_paths)

    @__not_empty_tree__.__get__(object)
    def __merge_reverse_index__(self, another_reverse_index: reverse_index_type, additional_path: List[int] = None):
        if additional_path is None:
            for element, paths in another_reverse_index.items():
                self.__add_paths__(element, paths)
        else:
            self.__merger_reverse_index_modified_paths__(another_reverse_index, additional_path)

    def __get_reverse_index__(self) -> reverse_index_type:
        return self._reverse_index

    def __get_root__(self) -> BaseNode:
        return self._root

    @BaseNode.__only_node_type__
    @__not_empty_tree__.__get__(object)
    def __child_indexing__(self, element: BaseNode, path_to_element: List[int]):
        for i in range(0, len(element)):
            child = element[i]
            if not child.empty():
                self.__child_indexing__(child, path_to_element + [i])
            self._reverse_index[child].append(path_to_element + [i])

    @BaseNode.__only_node_type__
    def __add_paths__(self, element: BaseNode, paths: List[List[int]]):
        self._counter += len(paths)
        self._reverse_index[element] += paths
        self.__last_inserted_path = paths[-1]

    @BaseNode.__only_node_type__
    def __add_path__(self, element: BaseNode, path: List[int], *args):
        path += args
        self._counter += 1
        self._reverse_index[element].append(path)
        self.__last_inserted_path = path

    def __str__(self):
        if self.empty():
            return ""
        else:
            return f"Root name: {self._root.name()}\nRoot kind: {self._root.kind()}\n--------------------"

    def empty(self):
        return len(self) == 0

    @BaseNode.__only_node_type__
    @__child_indexing_after__.__get__(object)
    def set_root(self, element: BaseNode) -> int:
        """
        Set root element.
        Return 1 if root was set.
        If root was existed, return 0.

        :param element: BaseNode instance
        :return: int
        """
        if self._root is None:
            self._root = element
            self.__add_path__(self._root, [])
            return 1

        return 0

    @__not_empty_tree__.__get__(object)
    def get(self, path: List[int]) -> Union[BaseNode, None]:
        """
        Get BaseNode instance from tree.

        Return None if element isn't in tree or path is wrong.

        :param path: List of integer index
        :return: BaseNode instance or None
        """
        node: BaseNode = self._root
        for ind in path:
            node = node[ind]
            if node is None:
                return None
        return node

    @__not_empty_tree__.__get__(object)
    @__child_indexing_after__.__get__(object)
    def add(self, element: BaseNode, path: List[int]) -> int:
        """
        Adding element to tree.

        Return -1 if path for adding is wrong.

        Return 0 is element with its name is already in tree.

        Else return 1

        :param path: list of indexes
        :param element: BaseNode instance
        :return: int
        """
        node: BaseNode = self._root
        for ind in path:
            node = node[ind]
            if node is None:
                return -1

        index = len(node)
        if node.add(element):
            self.__add_path__(element, path, index)
            return 1
        return 0

    @BaseNode.__only_node_type__
    def paths(self, element: BaseNode) -> List[List[int]]:
        """
        Return all known paths to element in tree

        :param element: BaseNode instance
        :return: list of paths to element
        """
        if element in self._reverse_index:
            return self._reverse_index[element]
        return []

    def paths_generator(self, element: BaseNode):
        """
        Generator for path returning

        :param element: BaseNode instance
        :return: one path on every step
        """
        for path in self.paths(element):
            yield path

    def paths_by_name(self, name: str) -> List[List[int]]:
        """
        Get paths to element by name

        :param name: string
        :return: list of paths
        """
        return self.paths(BaseNode(name, None))

    def paths_generator_by_name(self, name: str):
        """
        Generator for paths to element by name

        :param name: str
        :return: one path on every step
        """
        for path in self.paths(BaseNode(name, None)):
            yield path

    def show(self, target_file=None, **kwargs):
        """
        In depth first tree printing

        :param target_file: if none show in stdout. Else it uses target_file for outputting
        :return: None
        """
        if target_file is None:
            target_file = sys.stdout
        print(self, file=target_file)
        self._root.show(target_file=target_file, **kwargs)



from typing import Union as __Union__
from typing import Generator as __Generator__

from clangast.data_structures.tree import BaseTree as __Tree__
from clangast.data_structures.node import BaseNode as __Node__
from clangast.algorithms.__utility__ import (
    __not_none__, __cursor_kind__, __cursor_index__, __token_kind__
)

from typing import Any as __Any__

from functools import partial as __partial__

__left_borders__ = ('(', '[', '{', 'new_line')
__right_borders__ = (')', ']', '}', ';')

__borders_index__ = {
    **dict(zip(__left_borders__, __right_borders__)),
    **dict(zip(__right_borders__, __left_borders__))
}


class __MainGenerator__:
    def __init__(self, token_generator: __Generator__[__cursor_index__.Token, __Any__, None]):
        self.generator = token_generator
        self.tokens = []

    def __next__(self):
        for token in self.generator:
            while len(self.tokens) != 0:
                return self.tokens.pop()
            return token

    def __iter__(self):
        for t in self.generator:
            while len(self.tokens) != 0:
                yield self.tokens.pop()
            yield t

    def stack(self, value: __cursor_index__):
        self.tokens.append(value)


def __element_pretty_format__(element: __Union__[__cursor_index__.Cursor,
                                                 __cursor_index__.Token],
                              preamble: str = None) -> str:
    if preamble is not None:
        preamble = f"{preamble}\n"
    else:
        preamble = ""
    return f"{preamble}Name:\t{element.spelling}\nKind:\t{element.kind}\n----------------"


def __mismatch_spelling__(_token: __cursor_index__.Token, _spelling: str) -> bool:
    if _token is not None:
        return not _token.spelling == _spelling
    return True


def __parse_statement_stacked_generator__(token_generator: __MainGenerator__
                                          ) -> __Generator__[__Node__, __Any__, None]:
    # statement type: in brackets or single line
    token = next(token_generator)

    close_token = '}'

    if __mismatch_spelling__(token, '{'):
        close_token = ';'
        token_generator.stack(token)

    parser = __partial__(__parse_statement_to_stop_token__, close_token)
    for _node in filter(__not_none__,
                        parser(token_generator)):
        yield _node


def __parse_statement_to_stop_token__(stop_token_spelling: str,
                                      token_generator: __MainGenerator__) -> __Generator__[__Node__,
                                                                                           __Any__,
                                                                                           None]:
    for token in filter(__not_none__, token_generator):
        if token.spelling == stop_token_spelling:
            break
        node = __kind_switch__(token, token_generator)
        if node is not None:
            yield node
    yield None


def __gathering_tokens_between_borders__(token_generator: __MainGenerator__,
                                         close_token_spelling: str) -> __Generator__[__cursor_index__.Token,
                                                                                     __Any__,
                                                                                     None]:
    if close_token_spelling not in __borders_index__:
        return

    open_token_spelling = __borders_index__[close_token_spelling]
    inner_open_borders = 0

    for token in filter(__not_none__, token_generator):
        if token.spelling == open_token_spelling:
            inner_open_borders += 1

        if token.spelling == close_token_spelling:
            if inner_open_borders == 0:
                break
            else:
                inner_open_borders -= 1

        yield token


def __parse_id__(token: __cursor_index__.Token,
                 token_generator: __MainGenerator__
                 ) -> __Node__:
    node = __Node__(token.spelling, token.cursor.kind)
    node.store(
        something=" ".join((
            _token.spelling for _token in __gathering_tokens_between_borders__(token_generator, ';')
        ))
    )
    return node


def __parse_if_case__(token: __cursor_index__.Token,
                      token_generator: __MainGenerator__
                      ) -> __Node__:
    # set "if" as name for node
    node = __Node__(token.spelling, token.cursor.kind)

    # condition
    if __mismatch_spelling__(next(token_generator), '('):
        return node
    condition = " ".join(
        (token.spelling
         for token in __gathering_tokens_between_borders__(token_generator, ')'))
    )
    node.store(condition)

    # inner statement on if branch
    for _node in __parse_statement_stacked_generator__(token_generator):
        node.add(_node)

    # else
    token = next(token_generator)
    if __mismatch_spelling__(token, "else"):
        token_generator.stack(token)
        return node
    node.add(__parse_else__(token, token_generator))

    return node


def __parse_else__(token: __cursor_index__.Token,
                   tokens_generator: __MainGenerator__) -> __Node__:
    node = __Node__(token.spelling, token.cursor.kind)
    for _node in __parse_statement_stacked_generator__(tokens_generator):
        node.add(_node)
    return node


def __parse_switch_case__(token: __cursor_index__.Token,
                          token_generator: __MainGenerator__

                          ) -> __Node__:
    node = __Node__(token.spelling, token.cursor.kind)

    # condition
    if __mismatch_spelling__(next(token_generator), '('):
        return node
    node.store(
        something=" ".join((
            _token.spelling for _token in __gathering_tokens_between_borders__(token_generator, ')')
        ))
    )
    return node


def __parse_return__(token: __cursor_index__.Token,
                     token_generator: __MainGenerator__

                     ) -> __Node__:
    node = __Node__(token.spelling, token.cursor.kind)
    node.store(
        something=" ".join((
            token.spelling for token in __gathering_tokens_between_borders__(token_generator, ';')
        ))
    )
    return node


def __keyword_case__(token: __cursor_index__.Token,
                     token_generator: __MainGenerator__

                     ) -> __Union__[__Node__, None]:
    name = token.spelling
    if name == "if":
        return __parse_if_case__(token, token_generator)
    elif name == "switch":
        return __parse_switch_case__(token, token_generator)
    elif name == "return":
        return __parse_return__(token, token_generator)
    else:
        return None


def __id_case__(token: __cursor_index__.Token,
                token_generator: __MainGenerator__
                ) -> __Node__:
    return __parse_id__(token, token_generator)


def __kind_switch__(token: __cursor_index__.Token,
                    token_generator: __MainGenerator__
                    ) -> __Union__[__Node__, None]:
    kind = token.kind
    if kind == __token_kind__.IDENTIFIER:
        return __id_case__(token, token_generator)
    elif kind == __token_kind__.KEYWORD:
        return __keyword_case__(token, token_generator)
    else:
        return None


def __parse_statement_generator__(token_generator: __MainGenerator__) -> __Generator__[__Node__,
                                                                                       __Any__,
                                                                                       None]:
    for token in token_generator:
        node = __kind_switch__(token, token_generator)
        if node is not None:
            yield node
    yield None


def __parse_function_body__(main_generator: __MainGenerator__) -> __Generator__[__Node__, __Any__, None]:
    for node in __parse_statement_generator__(main_generator):
        yield node


def __parse_function__(cursor: __cursor_index__.Cursor) -> __Tree__:
    main_generator = __MainGenerator__(cursor.get_tokens())

    root_node = __Node__(cursor.spelling, cursor.kind)
    root_node.store(
        something=" ".join((
            token.spelling for token in __gathering_tokens_between_borders__(main_generator, '{')
        ))
    )

    tree = __Tree__(root=root_node)
    for node in filter(__not_none__, __parse_function_body__(main_generator)):
        tree.add(node, [])
    return tree



from clang import cindex
from typing import List, Dict, Hashable, Any
from functools import wraps

import sys


class BaseNode(object):
    index_type = List[Any]
    reverse_index_type = Dict[Hashable, int]

    @staticmethod
    def __first_argument_is_string__(method):
        @wraps(method)
        def __check__(self, arg, *args, **kwargs):
            if type(arg) != str:
                raise TypeError(f"First argument must be string; type of first element: {type(arg)}")
            return method(self, arg, *args, **kwargs)

        return __check__

    @__first_argument_is_string__.__get__(object)
    def __init__(self, spelling: str, kind: cindex.CursorKind):
        """
        Base class for trees nodes

        :param spelling: Name of element; str
        :param kind: kind of element; clang.cindex.CursorKind
        """
        self._index: index_type = []
        self._reverse_index: reverse_index_type = {}

        self._name: str = spelling
        self._kind: cindex.CursorKind = kind

        self._additional = list()

        self._pretty_print_offset = ""

    @staticmethod
    def __node_type__(element):
        return type(element) == BaseNode

    @staticmethod
    def __only_node_type__(method):
        @wraps(method)
        def __check_type__(self, element, *args, **kwargs):
            if not BaseNode.__node_type__(element):
                raise TypeError(f"Must be Node type; type of element: {type(element)}")
            return method(self, element, *args, **kwargs)

        return __check_type__

    def __hash__(self):
        return self._name.__hash__()

    def __next__(self):
        for something in self._additional:
            yield something

    @__only_node_type__.__get__(object)
    def __eq__(self, element):
        return self._name == element.name()

    def __ne__(self, element):
        return not (self.__eq__(element))

    def __getitem__(self, key: int):
        """
        If index in child return child by index.
        Else return None

        :param key: int
        :return: BaseNode instance or None
        """
        if 0 <= key < len(self):
            return self._index[key]
        return None

    def __len__(self):
        return len(self._index)

    def __str__(self):
        self_description = f"{self._pretty_print_offset}Kind:\t{self._kind}\n{self._pretty_print_offset}Name:\t{self._name}\n"
        end_line = f"{self._pretty_print_offset}--------------"
        storage = self._pretty_print_offset + f"\n{self._pretty_print_offset}".join(
            (str(store_element) for store_element in self._additional))
        storage = f"{self._pretty_print_offset}Store:\n{storage}\n"
        return f"{self_description}{storage}{end_line}"

    def empty(self) -> bool:
        """
        Return true is node has not child

        :return: bool
        """
        return len(self) == 0

    def name(self) -> str:
        """
        Get name of element

        :return: string
        """
        return self._name

    def kind(self) -> cindex.CursorKind:
        """
        Get cursor kind of element

        :return: clang.cindex.CursorKind
        """
        return self._kind

    @__only_node_type__.__get__(object)
    def has(self, element) -> bool:
        """
        Return True if element has in children of elements

        :param element: BaseNode instance
        :return: bool
        """
        return element in self._reverse_index

    def has_by_name(self, name: str) -> bool:
        """
        Return True if element with name in child of elements

        :param name: string
        :return: bool
        """
        return self.has(BaseNode(name, None))

    @__only_node_type__.__get__(object)
    def add(self, element) -> bool:
        """
        Adding new element to child if it isn't in.
        Return true if adding has finished success.

        :param element: BaseNode instance
        :return: bool
        """
        if self.has(element):
            return False

        index = len(self)
        self._index.append(element)
        self._reverse_index[element] = index

        return True

    @__only_node_type__.__get__(object)
    def index(self, element) -> int:
        """
        Get integer index of element

        :param element: BaseNode instance
        :return: int
        """
        if self.has(element):
            return self._reverse_index[element]
        return len(self)

    def index_by_name(self, name: str) -> int:
        """
        Get integer index of element by name. Searching uses only name.
        Useful with get item by []

        :param name: string
        :return: int
        """
        return self.index(BaseNode(name, None))

    def store(self, something):
        """
        Store something to additional storage of node

        :param something: Any
        :return: None
        """
        self._additional.append(something)

    def get_from_store(self, index: int) -> Any:
        """
        Get element from additional storage element by index.
        For iterating over this storage better use iterating over node with `for <> in <>`.

        :param index: int
        :return: Any if element in storage. None another.
        """
        if index < len(self._additional):
            return self._additional[index]
        return None

    def show(self, offset: str = None, offset_block: str = None,
             target_file=None):
        """
        Pretty output current node info and its childs info with offset

        :param offset: symbols before line, default is 2 space
        :param offset_block: step for offset increasing, default is 2 space
        :param target_file: descriptor for printing, default is stdout
        :return: None
        """
        prev_offset = self._pretty_print_offset
        if offset is None:
            offset = prev_offset
        self._pretty_print_offset = offset

        if offset_block is None:
            offset_block = "  "

        if target_file is None:
            target_file = sys.stdout

        print(self, file=target_file)
        print(f"{offset}[", file=target_file)
        for i in range(0, len(self)):
            self._index[i].show(offset + offset_block, offset_block, target_file)
        print(f"{offset}]", file=target_file)

        self._pretty_print_offset = prev_offset



import random as __random__
import datetime as __time__

from functools import partial as __partial__

__open_number__ = set()


def __get_debug_id__(operation_name: str) -> str:
    global __open_number__

    random_number = __random__.randint(0, 1000)
    while random_number in __open_number__:
        __random__.seed(random_number)
        random_number = __random__.randint(0, 100000)

    return f"{operation_name}_{random_number}"


def __debug_message__(_id: str, message: str) -> None:
    current_time = str(__time__.datetime.now().time())
    print(f"{current_time}::{_id}::{message}")


def __get_local_debug_logger__(operation_type: str):
    _id = __get_debug_id__(operation_type)
    return __partial__(__debug_message__, _id)



from clang import cindex as __cursor_index__

__token_kind__ = __cursor_index__.TokenKind
__cursor_kind__ = __cursor_index__.CursorKind

from typing import Any as __Any__

__CLANG_INIT__ = False

from typing import List as __List__


def __init_clang__(_path_to_lib: str):
    global __CLANG_INIT__
    if not __CLANG_INIT__:
        __cursor_index__.Config.set_library_file(_path_to_lib)
        __CLANG_INIT__ = True


def __not_none__(_any: __Any__) -> bool:
    return _any is not None


def __all_as_kinds_generator__(kinds: __List__[__cursor_index__.CursorKind], cursor: __cursor_index__.Cursor):
    for element in list(cursor.get_children())[::-1]:
        if element.kind in kinds:
            yield element
